//  LauncherOSX
//
//  Created by Boris Schneiderman.
//  Copyright (c) 2014 Readium Foundation and/or its licensees. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this 
//  list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation and/or 
//  other materials provided with the distribution.
//  3. Neither the name of the organization nor the names of its contributors may be 
//  used to endorse or promote products derived from this software without specific 
//  prior written permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
//  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
//  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//  OF THE POSSIBILITY OF SUCH DAMAGE.

ReadiumSDK.Models.Switches = function() {

};

// Description: Parse the epub "switch" tags and hide
// cases that are not supported
ReadiumSDK.Models.Switches.apply = function(dom) {


    // helper method, returns true if a given case node
    // is supported, false otherwise
    function isSupported(caseNode) {

        var ns = caseNode.attributes["required-namespace"];
        if(!ns) {
            // the namespace was not specified, that should
            // never happen, we don't support it then
            console.log("Encountered a case statement with no required-namespace");
            return false;
        }
        // all the xmlns that readium is known to support
        // TODO this is going to require maintenance
        var supportedNamespaces = ["http://www.w3.org/1998/Math/MathML"];
        return _.include(supportedNamespaces, ns.nodeValue); // USE_ADOBE_DRM. On mac, we should use ns.nodeValue
    }

    $('epub\\:switch,switch', dom).each( function() { // USE_ADOBE_DRM. IE does not understand just 'switch'

        // keep track of whether or now we found one
        var found = false;

        $('epub\\:case,case', this).each(function() { // USE_ADOBE_DRM. IE does not understand just 'case'

            if( !found && isSupported(this) ) {
                found = true; // we found the node, don't remove it
            }
            else {
                $(this).remove(); // remove the node from the dom
//                    $(this).prop("hidden", true);
            }
        });

        if(found) {
            // if we found a supported case, remove the default
            $('epub\\:default,default', this).remove(); // USE_ADOBE_DRM. IE does not understand just 'default'
//                $('default', this).prop("hidden", true);
        }
    })
    
    $('form', dom).each( function() { // USE_ADOBE_DRM

        this.addEventListener('submit', function(e) {
            e.currentTarget.classList.add('submitted');

            for (i = 0; i < this.elements.length; i++) {
                var val = this.elements[i].value;
                if (this.elements[i].type == "tel") {
                    var telPattern = this.elements[i].pattern;
                    if (telPattern == '') {
                        telPattern = "/^\(?([0-9]{3})\)?[-. ]?([0-9]{3})[-. ]?([0-9]{4})$/";
                    }
                    if (val == '' || !val.match(telPattern)) {
                        this.elements[i].style.borderColor="red";
                        return false;
                    }
                    else {
                        this.elements[i].style.borderColor="green";
                    }
                }
                else if (this.elements[i].type == "url") {
                    var urlPattern = /(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%@!\-\/]))?/;
                    if (val == '' || !urlPattern.test(val)) {
                        this.elements[i].style.borderColor="red";
                        return false;
                    }
                    else {
                        this.elements[i].style.borderColor="green";
                    }
                }
                else if (this.elements[i].type == "email") {
                    var re = /[A-Z0-9._%+-]+@[A-Z0-9.-]+.[A-Z]{2,4}/igm;
                    if (val == '' || !re.test(val)) {
                        this.elements[i].style.borderColor="red";
                        return false;
                    }
                    else {
                        this.elements[i].style.borderColor="green";
                    }
                }
                else if (this.elements[i].type == "number") {
                    if (val == '' || isNaN(parseFloat(val))) {
                        this.elements[i].style.borderColor="red";
                        return false;
                    }
                    else {
                        this.elements[i].style.borderColor="green";
                    }
                }
            }
            return true;
        });
    }) // USE_ADOBE_DRM
    
    $('input', dom).each( function() { // USE_ADOBE_DRM
        if (this.attributes["type"].value == "number") {
            this.addEventListener('keyup', function(e) {
                val = this.value;
                if ( isNaN(parseFloat(val))) {
                    this.style.borderColor="red";
                }
                else {
                    this.style.borderColor="green";
                }
            });
        }
    }) // USE_ADOBE_DRM
    
};
