
//  LauncherOSX
//
//  Created by Boris Schneiderman.
//  Copyright (c) 2014 Readium Foundation and/or its licensees. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this 
//  list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation and/or 
//  other materials provided with the distribution.
//  3. Neither the name of the organization nor the names of its contributors may be 
//  used to endorse or promote products derived from this software without specific 
//  prior written permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
//  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
//  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//  OF THE POSSIBILITY OF SUCH DAMAGE.

/*
 * Renders reflowable content using CSS columns
 *
 * @class ReadiumSDK.Views.ReflowableView
 */

ReadiumSDK.Views.ReflowableView = function(options){

    _.extend(this, Backbone.Events);

    var self = this;
    
    var _$viewport = options.$viewport;
    var _spine = options.spine;
    var _userStyles = options.userStyles;
    var _bookStyles = options.bookStyles;
    var _iframeLoader = options.iframeLoader;
    
    var _currentSpineItem;
    var _isWaitingFrameRender = false;    
    var _deferredPageRequest;
    var _fontSize = 100;
    var _$contentFrame;
    var _navigationLogic;
    var _$el;
    var _$iframe;
    var _$epubHtml;
    
    var _$htmlBody;
    
    var _htmlBodyIsVerticalWritingMode;
    var _htmlBodyIsLTRDirection;
    var _htmlBodyIsLTRWritingMode;
    
    
    var _currentOpacity = -1;

    var _lastViewPortSize = {
        width: undefined,
        height: undefined
    };

    var _paginationInfo = {

        visibleColumnCount : 2,
        columnGap : 20,
        spreadCount : 0,
        currentSpreadIndex : 0,
        columnWidth : undefined,
        pageOffset : 0,
        columnCount: 0
    };

    this.render = function(){

        var template = ReadiumSDK.Helpers.loadTemplate("reflowable_book_frame", {});

        _$el = $(template);
        _$viewport.append(_$el);
        
        // This fixes rendering issues with WebView (native apps), probably related to hardware acceleration
        _$el.css("transform", "translateZ(0)");

        var settings = _viewSettings;
        if (!settings)
        {
            //defaults
            settings = new ReadiumSDK.Models.ViewerSettings({});
        }
        if (settings.enableGPUHardwareAccelerationCSS3D) {
            // This fixes rendering issues with WebView (native apps), which clips content embedded in iframes unless GPU hardware acceleration is enabled for CSS rendering.
            _$el.css("transform", "translateZ(0)");
        }

        // See ReaderView.handleViewportResize
        // var lazyResize = _.debounce(self.onViewportResize, 100);
        // $(window).on("resize.ReadiumSDK.reflowableView", _.bind(lazyResize, self));
        renderIframe();

        return self;
    };

    function setFrameSizesToRectangle(rectangle) {
        _$contentFrame.css("left", rectangle.left + "px");
        _$contentFrame.css("top", rectangle.top + "px");
        _$contentFrame.css("right", rectangle.right + "px");
        _$contentFrame.css("bottom", rectangle.bottom + "px");

    }

    this.remove = function() {

        //$(window).off("resize.ReadiumSDK.reflowableView");
        _$el.remove();

    };

    this.isReflowable = function() {
        return true;
    };

    this.onViewportResize = function() {

        if(updateViewportSize()) {
            //depends on aspect ratio of viewport and rendition:spread-* setting we may have to switch spread on/off
            updateColumnCount();
            updatePagination();
        }

        /**USE_ADOBE_DRM we need to force rerendering on resize for RTL content to display properly
         * Bug#4039974: This causes issues with high speed devices. Disabling for Android.
         */
        var delay = ReadiumSDK._Android ? 1 : 50;
		_$epubHtml.hide();
            setTimeout(function() {
                _$epubHtml.show();
                onPaginationChanged(self); // => redraw() => showBook()
         }, delay);

    };

    var _viewSettings = undefined;
    this.setViewSettings = function(settings) {
        
        _viewSettings = settings;

        _paginationInfo.columnGap = settings.columnGap;
        _fontSize = settings.fontSize;

        updateHtmlFontSize();
        updateColumnGap();
        updateTextAlignToJustify(); //USE_ADOBE_DRM
        
        updateViewportSize();
        updateColumnCount();
        updatePagination();
    };

    function updateColumnCount() {
        _paginationInfo.visibleColumnCount = _htmlBodyIsVerticalWritingMode ? 1 : (ReadiumSDK.Helpers.deduceSyntheticSpread(_$viewport, _currentSpineItem, _viewSettings) ? 2 : 1);
    }

    function renderIframe() {
        if (_$contentFrame) {
            //destroy old contentFrame
            _$contentFrame.remove();
        }

        var template = ReadiumSDK.Helpers.loadTemplate("reflowable_book_page_frame", {});
        var $bookFrame = $(template);
        $bookFrame = _$el.append($bookFrame);

        _$contentFrame = $("#reflowable-content-frame", $bookFrame);

        _$iframe = $("#epubContentIframe", $bookFrame);

        _$iframe.css("left", "");
        _$iframe.css("right", "");
        _$iframe.css("position", "relative");
        //_$iframe.css(_spine.isLeftToRight() ? "left" : "right", "0px");
        _$iframe.css("overflow", "hidden");

        _navigationLogic = new ReadiumSDK.Views.CfiNavigationLogic(
            _$contentFrame, _$iframe,
            { rectangleBased: true, paginationInfo: _paginationInfo });
    }

    function loadSpineItem(spineItem) {

        if(_currentSpineItem != spineItem) {

            //create & append iframe to container frame
            renderIframe();

            _paginationInfo.pageOffset = 0;
            _paginationInfo.currentSpreadIndex = 0;
            _currentSpineItem = spineItem;
            _isWaitingFrameRender = true;

            var src = _spine.package.resolveRelativeUrl(spineItem.href);
            self.trigger(ReadiumSDK.Events.CONTENT_DOCUMENT_LOAD_START, _$iframe, spineItem);

            _$iframe.css("opacity", "0.01");
            
            _iframeLoader.loadIframe(_$iframe[0], src, onIFrameLoad, self, {spineItem : spineItem});
        }
    }

    function updateHtmlFontSize() {

        if(_$epubHtml) {
            _$epubHtml.css("font-size", _fontSize + "%");
        }
    }

    function updateColumnGap() {

        if(_$epubHtml) {
        
            _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
                _$epubHtml.css(prefix + "column-gap", _paginationInfo.columnGap + "px");
            });
        }
    }

    //USE_ADOBE_DRM
    function updateTextAlignToJustify()
    {
        if(_$epubHtml) {
            _$epubHtml.css("text-align", "justify");
        }        
    }

    function onIFrameLoad(success) {

        _isWaitingFrameRender = false;

        //while we where loading frame new request came
        if(_deferredPageRequest && _deferredPageRequest.spineItem != _currentSpineItem) {
            loadSpineItem(_deferredPageRequest.spineItem);
            return;
        }

        if(!success) {
            _$iframe.css("opacity", "1");
            _deferredPageRequest = undefined;
            return;
        }

        self.trigger(ReadiumSDK.Events.CONTENT_DOCUMENT_LOADED, _$iframe, _currentSpineItem);

        var epubContentDocument = _$iframe[0].contentDocument;
        _$epubHtml = $("html", epubContentDocument);
        _$htmlBody = $("body", _$epubHtml);
        
        _htmlBodyIsVerticalWritingMode = false;
        _htmlBodyIsLTRDirection = true;
        _htmlBodyIsLTRWritingMode = undefined;
        
        var win = _$iframe[0].contentDocument.defaultView || _$iframe[0].contentWindow;
        
        //Helpers.isIframeAlive
        var htmlBodyComputedStyle = win.getComputedStyle(_$htmlBody[0], null);
        if (htmlBodyComputedStyle)
        {
            _htmlBodyIsLTRDirection = htmlBodyComputedStyle.direction === "ltr";

            var writingMode = undefined;
            if (htmlBodyComputedStyle.getPropertyValue)
            {
                writingMode = htmlBodyComputedStyle.getPropertyValue("-webkit-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-moz-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-ms-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-o-writing-mode") || htmlBodyComputedStyle.getPropertyValue("-epub-writing-mode") || htmlBodyComputedStyle.getPropertyValue("writing-mode");
            }
            else
            {
                writingMode = htmlBodyComputedStyle.webkitWritingMode || htmlBodyComputedStyle.mozWritingMode || htmlBodyComputedStyle.msWritingMode || htmlBodyComputedStyle.oWritingMode || htmlBodyComputedStyle.epubWritingMode || htmlBodyComputedStyle.writingMode;
            }

            if (writingMode)
            {
                _htmlBodyIsLTRWritingMode = writingMode.indexOf("-lr") >= 0;
            
                if (writingMode.indexOf("vertical") >= 0 || writingMode.indexOf("tb-") >= 0 || writingMode.indexOf("bt-") >= 0)
                {
                    _htmlBodyIsVerticalWritingMode = true;
                }
            }
        }

        if (_htmlBodyIsLTRDirection)
        {
            if (_$htmlBody[0].getAttribute("dir") === "rtl" || _$epubHtml[0].getAttribute("dir") === "rtl")
            {
                _htmlBodyIsLTRDirection = false;
            }
        }

        // USE_ADOBE_DRM. 
        // Without this, epubtest-0130.epub's front.xhtml renders wrongly.
        // Some EPUBs may not have explicit RTL content direction (via CSS "direction" property or @dir attribute) despite having a RTL page progression direction. Readium consequently tweaks the HTML in order to restore the correct block flow in the browser renderer, resulting in the appropriate CSS columnisation (which is used to emulate pagination).
        if (!_spine.isLeftToRight() && _htmlBodyIsLTRDirection && (!_htmlBodyIsVerticalWritingMode || ReadiumSDK._isIEBrowser))
        {
            _$htmlBody[0].setAttribute("dir", "rtl");
            _htmlBodyIsLTRDirection = false;
            _htmlBodyIsLTRWritingMode = false;
        }        
        
        _paginationInfo.isVerticalWritingMode = _htmlBodyIsVerticalWritingMode;
        
        hideBook();
        _$iframe.css("opacity", "1");
        
        updateViewportSize();
        _$epubHtml.css("height", _lastViewPortSize.height + "px");
        
        _$epubHtml.css("position", "relative");
        _$epubHtml.css("margin", "0");
        _$epubHtml.css("padding", "0");

        _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
            _$epubHtml.css(prefix + "column-axis", (_htmlBodyIsVerticalWritingMode ? "vertical" : "horizontal"));
        });

        //
        // /////////
        // //Columns Debugging
        //
        // _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
        //     _$epubHtml.css(prefix + "column-rule-color", "red");
        //     _$epubHtml.css(prefix + "column-rule-style", "dashed");
        //     _$epubHtml.css(prefix + "column-rule-width", "1px");
        // });
        // _$epubHtml.css("background-color", '#b0c4de');
        //
        // ////
        
        self.applyBookStyles();
        resizeImages();

        updateHtmlFontSize();
        updateColumnGap();
        updateTextAlignToJustify(); //USE_ADOBE_DRM


        self.applyStyles();
    }

    this.applyStyles = function() {

        ReadiumSDK.Helpers.setStyles(_userStyles.getStyles(), _$el.parent());

        //because left, top, bottom, right setting ignores padding of parent container
        //we have to take it to account manually
        var elementMargins = ReadiumSDK.Helpers.Margins.fromElement(_$el);
        setFrameSizesToRectangle(elementMargins.padding);


        updateViewportSize();
        updateColumnCount();
        updatePagination();

    };

    this.applyBookStyles = function() {

        if(_$epubHtml) {
            ReadiumSDK.Helpers.setStyles(_bookStyles.getStyles(), _$epubHtml);
        }
    };

    function openDeferredElement() {

        if(!_deferredPageRequest) {
            return;
        }

        var deferredData = _deferredPageRequest;
        _deferredPageRequest = undefined;
        self.openPage(deferredData);

    }

	//USE_ADOBE_DRM start. Added for test cases epubcfi-030, 040, 050
    this.LoadVideoAtTemporalOffset = function (pageRequest) {
        var pos = pageRequest.elementCfi.indexOf('~');
        var end = pageRequest.elementCfi.indexOf('@');
        if (end < 0)
            end = pageRequest.elementCfi.length;
        if (pos > 0) {
            var toffset = pageRequest.elementCfi.slice(pos + 1, end);
            $element = this.getElementByCfi(pageRequest.spineItem, pageRequest.elementCfi);
            var ele = $element[0];
            var name = ele.nodeName ? ele.nodeName.toLowerCase() : undefined;
            if (name === "audio" || name === "video") {
                if (ele.readyState > 0) {
                    ele.currentTime = toffset;
                    ele.play();
                }
            }
        }

    }
	//USE_ADOBE_DRM end
    this.openPage = function(pageRequest) {

        if(_isWaitingFrameRender) {
            _deferredPageRequest = pageRequest;
            return;
        }

		//USE_ADOBE_DRM
        try{
		  onLoadOfIFrame(document.getElementById('epubContentIframe'));
//                                alert("onLoadOfIFrame");
        }catch(error){}

        // if no spine item specified we are talking about current spine item
        if(pageRequest.spineItem && pageRequest.spineItem != _currentSpineItem) {
            _deferredPageRequest = pageRequest;
            loadSpineItem(pageRequest.spineItem);
            return;
        }
        else {
            self.trigger(ReadiumSDK.InternalEvents.CURRENT_VIEW_NEW_PAGE_READY);
        }

        var pageIndex = undefined;


        if(pageRequest.spineItemPageIndex !== undefined) {
            pageIndex = pageRequest.spineItemPageIndex;
        }
        else if(pageRequest.elementId) {
            pageIndex = _navigationLogic.getPageForElementId(pageRequest.elementId);
        }
        else if(pageRequest.elementCfi) {
            try
            {
                pageIndex = _navigationLogic.getPageForElementCfi(pageRequest.elementCfi,
                    ["cfi-marker", "mo-cfi-highlight", "MathJax_Preview", "MathJax_SVG_Display"],
                    [],
                    ["MathJax_Message"]);
            }
            catch (e)
            {
                pageIndex = 0;
                console.error(e);
            }
        }
        else if(pageRequest.firstPage) {
            pageIndex = 0;
        }
        else if(pageRequest.lastPage) {
            pageIndex = _paginationInfo.columnCount - 1;
        }
        else {
            console.debug("No criteria in pageRequest");
            pageIndex = 0;
        }

        if(pageIndex >= 0 && pageIndex < _paginationInfo.columnCount) {
			//USE_ADOBE_DRM start. Added for fixing epubcfi-030, 040, 050
            _paginationInfo.currentSpreadIndex = Math.floor(pageIndex / _paginationInfo.visibleColumnCount);
            if (pageRequest.elementCfi &&  !ReadiumSDK._iOS)
            {
                $element = this.getElementByCfi(pageRequest.spineItem, pageRequest.elementCfi);
                var ele = $element[0];
                if (ele.readyState > 0)
                    this.LoadVideoAtTemporalOffset(pageRequest);
                else
                    ele.addEventListener("canplay", function () { this.LoadVideoAtTemporalOffset(pageRequest); });
            }
            // USE_ADOBE_DRM end
            
            onPaginationChanged(pageRequest.initiator, pageRequest.spineItem, pageRequest.elementId);
        }
        else {
            console.log('Illegal pageIndex value: ', pageIndex, 'column count is ', _paginationInfo.columnCount);
        }
    };

    function redraw() {

        var offsetVal =  -_paginationInfo.pageOffset + "px";

        /**USE_ADOBE_DRM
         * #IEVerticalMode layout means we do manual width-based
         * columnization for consistency. Hence use the left/right tags */
        if (_htmlBodyIsVerticalWritingMode && !ReadiumSDK._isIEBrowser)
        {
            _$epubHtml.css("top", offsetVal);
        }
        else
        {
            var ltr = _htmlBodyIsLTRDirection || _htmlBodyIsLTRWritingMode;

            _$epubHtml.css("left", ltr ? offsetVal : "");
            _$epubHtml.css("right", !ltr ? offsetVal : "");
        }

        showBook(); // as it's no longer hidden by shifting the position
    }

    function updateViewportSize() {

        var newWidth = _$contentFrame.width()-(_viewSettings.getMarginLeftCSSPx()+_viewSettings.getMarginRightCSSPx()); // USE_ADOBE_DRM. Added to support margins
        var newHeight = _$contentFrame.height()-(_viewSettings.getMarginTopCSSPx()+_viewSettings.getMarginBottomCSSPx()); // USE_ADOBE_DRM

        // USE_ADOBE_DRM
        if(newWidth < 0 || newHeight < 0) {
            newWidth = newHeight = 0;
        }

        if(_lastViewPortSize.width !== newWidth || _lastViewPortSize.height !== newHeight){

            _lastViewPortSize.width = newWidth;
            _lastViewPortSize.height = newHeight;
            return true;
        }

        return false;
    }

    function onPaginationChanged(initiator, paginationRequest_spineItem, paginationRequest_elementId) {

        _paginationInfo.pageOffset = (_paginationInfo.columnWidth + _paginationInfo.columnGap) * _paginationInfo.visibleColumnCount * _paginationInfo.currentSpreadIndex;
        //USE_ADOBE_DRM start
        var info = AdobeRDMHelper.navigateToNthScreenFromCurrent_Info;
        if(info)
        {
            AdobeRDMHelper.navigateToNthScreenFromCurrent(info.targetPageIndex);
            return;
        }
        //USE_ADOBE_DRM end
        redraw();
        self.trigger(ReadiumSDK.InternalEvents.CURRENT_VIEW_PAGINATION_CHANGED, { paginationInfo: self.getPaginationInfo(), initiator: initiator, spineItem: paginationRequest_spineItem, elementId: paginationRequest_elementId } );
    }

    this.openPagePrev = function (initiator) {

        if(!_currentSpineItem) {
            return;
        }

        if(_paginationInfo.currentSpreadIndex > 0) {
            _paginationInfo.currentSpreadIndex--;
            onPaginationChanged(initiator);
        }
        else {

            var prevSpineItem = _spine.prevItem(_currentSpineItem, true);
            if(prevSpineItem) {

                var pageRequest = new ReadiumSDK.Models.PageOpenRequest(prevSpineItem, initiator);
                pageRequest.setLastPage();
                self.openPage(pageRequest);
            }
        }
    };

    this.openPageNext = function (initiator) {

        if(!_currentSpineItem) {
            return;
        }

        if(_paginationInfo.currentSpreadIndex < _paginationInfo.spreadCount - 1) {
            _paginationInfo.currentSpreadIndex++;
            onPaginationChanged(initiator);
        }
        else {

            var nextSpineItem = _spine.nextItem(_currentSpineItem, true);
            if(nextSpineItem) {

                var pageRequest = new ReadiumSDK.Models.PageOpenRequest(nextSpineItem, initiator);
                pageRequest.setFirstPage();
                self.openPage(pageRequest);
            }
        }
    };

    //USE_ADOBE_DRM
    this.openPageWithPageNumber = function (pageNumber,initiator){
        if(!_currentSpineItem) {
            return;
        }

        if( pageNumber < _paginationInfo.currentSpreadIndex )
            AdobeRDMHelper.ScreenInfoCache.CSIS.readingDir = AdobeRDMHelper.ScreenInfoCache.ReadingDir.BACKWARD;
        else
            AdobeRDMHelper.ScreenInfoCache.CSIS.readingDir = AdobeRDMHelper.ScreenInfoCache.ReadingDir.FORWARD;
        
        if(pageNumber < _paginationInfo.spreadCount) {
            _paginationInfo.currentSpreadIndex = pageNumber;
            onPaginationChanged(initiator);
        }
    };

    //USE_ADOBE_DRM start
    this.navigateToNthScreenFromCurrent = function(n, initiator){
        if(!_currentSpineItem) {
            return;
        }

        var targetPageIndex = _paginationInfo.currentSpreadIndex+n;
        if(targetPageIndex < 0)
        {
            targetPageIndex = targetPageIndex-_paginationInfo.currentSpreadIndex+1;            
            var prevSpineItem = _spine.prevItem(_currentSpineItem, true);
            if(prevSpineItem)
            {
                AdobeRDMHelper.navigateToNthScreenFromCurrent_Info = {targetPageIndex: targetPageIndex};
                this.openPagePrev(initiator);                
            }
            else
            {
                // We can't go more previous. Just stop jumping further and go to start of current spine item and notify.
                AdobeRDMHelper.navigateToNthScreenFromCurrent_Info = null;
                this.openPageWithPageNumber(0, initiator); 
            }
        }
        else if(targetPageIndex >= _paginationInfo.spreadCount)
        {
            targetPageIndex = targetPageIndex-_paginationInfo.spreadCount;
            var nextSpineItem = _spine.nextItem(_currentSpineItem, true);
            if(nextSpineItem)
            {
                AdobeRDMHelper.navigateToNthScreenFromCurrent_Info = {targetPageIndex: targetPageIndex};
                this.openPageNext(initiator);                                
            }
            else
            {
                // We are already on the final spine item. Just stop jumping further and go to end of current spine item and notify.
                AdobeRDMHelper.navigateToNthScreenFromCurrent_Info = null;
                targetPageIndex = _paginationInfo.spreadCount-1;
                this.openPageWithPageNumber(targetPageIndex, initiator);                
            }
        }
        else
        {
            AdobeRDMHelper.navigateToNthScreenFromCurrent_Info = null;
            this.openPageWithPageNumber(targetPageIndex, initiator);
        }
    };
    //USE_ADOBE_DRM end

    function updatePagination() {

        if(!_$epubHtml) {
            return;
        }

        hideBook(); // shiftBookOfScreen();

        _$iframe.css("left", _viewSettings.getMarginLeftCSSPx()); //USE_ADOBE_DRM
        _$iframe.css("top", _viewSettings.getMarginTopCSSPx()); //USE_ADOBE_DRM
        _$iframe.css("width", _lastViewPortSize.width + "px");
        _$iframe.css("height", _lastViewPortSize.height + "px");

        _$epubHtml.css("height", _lastViewPortSize.height + "px");
        _$epubHtml.css("min-height", _lastViewPortSize.height + "px");
        _$epubHtml.css("max-height", _lastViewPortSize.height + "px");

        //normalise spacing to avoid interference with column-isation
        _$epubHtml.css('margin', 0);
        _$epubHtml.css('padding', 0);
        _$epubHtml.css('border', 0);
        _$htmlBody.css('margin-bottom', 0);
        _$htmlBody.css('padding-bottom', 0);

        var spacing = 0;
        try
        {
            spacing = parseInt(_$htmlBody.css('padding-top')) + parseInt(_$htmlBody.css('border-top-width')) + parseInt(_$htmlBody.css('border-bottom-width'));
        }
        catch(err)
        {
            
        }

    
        //USE_ADOBE_DRM //calculate new viewport size by substracting the margin, this will be used to calculate the size for image wrapper elements.
        var win = _$iframe[0].contentDocument.defaultView || _$iframe[0].contentWindow;
        var computedStyle = win.getComputedStyle(_$htmlBody[0], null)
        var heightMargin = parseInt(computedStyle.marginTop,10) + parseInt(computedStyle.marginBottom,10);
        var widthMargin  = parseInt(computedStyle.marginLeft,10) + parseInt(computedStyle.marginRight,10);
        var bodySize = {width:_lastViewPortSize.width-widthMargin,height:_lastViewPortSize.height-heightMargin};


        // Needed for Firefox, otherwise content shrinks vertically, resulting in scrollWidth accomodating more columns than necessary
        //_$htmlBody.css("min-height", _lastViewPortSize.height-spacing-9 + "px");
        // USE_ADOBE_DRM. Commenting out the following 2 lines as they reset scrollHeight of the html to max-height, causing data 
        // loss(truncated chapters) on mac 10.8.
        //_$htmlBody.css("min-height", "50%");
        //_$htmlBody.css("max-height", bodySize.width//USE_ADOBE_DRM + "px");

        _paginationInfo.rightToLeft = _spine.isRightToLeft();

        /**USE_ADOBE_DRM #IEVerticalMode
         * IE's vertical writing mode behaves by laying out complete top-bottom
         * and then flowing left-right, which means columnization has to be
         * width based, not height based like in Chrome */
        // _paginationInfo.columnWidth = Math.round(((_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.height : _lastViewPortSize.width) - _paginationInfo.columnGap * (_paginationInfo.visibleColumnCount - 1)) / _paginationInfo.visibleColumnCount);
        _paginationInfo.columnWidth = Math.round(((!_htmlBodyIsVerticalWritingMode || ReadiumSDK._isIEBrowser ? _lastViewPortSize.width : _lastViewPortSize.height) - _paginationInfo.columnGap * (_paginationInfo.visibleColumnCount - 1)) / _paginationInfo.visibleColumnCount);

        _$epubHtml.css("width", (_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.width : _paginationInfo.columnWidth) + "px");
        _$epubHtml.css("min-width", (_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.width : _paginationInfo.columnWidth) + "px");
        _$epubHtml.css("max-width", (_htmlBodyIsVerticalWritingMode ? _lastViewPortSize.width : _paginationInfo.columnWidth) + "px");

        //updateImageWrapperSize({width:bodySize.width, height:bodySize.height});//USE_ADOBE_DRM

        /**USE_ADOBE_DRM START #IEVerticalMode
         * Setting column-width on IE with vertical writing mode
         * causes height to be auto-set to be viewport width, and removes reflow*/
        if(!(_htmlBodyIsVerticalWritingMode && ReadiumSDK._isIEBrowser)) {
            _.each(['-webkit-', '-moz-', '-ms-', ''], function(prefix) {
                _$epubHtml.css(prefix + "column-width", _paginationInfo.columnWidth + "px");
            });
        }
        //USE_ADOBE_DRM END

        _$epubHtml.css({left: "0", right: "0", top: "0"});
        
        ReadiumSDK.Helpers.triggerLayout(_$iframe);

        _paginationInfo.columnCount = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) + _paginationInfo.columnGap) / (_paginationInfo.columnWidth + _paginationInfo.columnGap);
        /**USE_ADOBE_DRM
         * Webkit seems to return round multiples of width in scrollWidth
         * IE returns fractions, so Math.round causes truncation. Math.ceil is the robust alternative*/
        _paginationInfo.columnCount = Math.ceil(_paginationInfo.columnCount);

        var totalGaps = (_paginationInfo.columnCount-1) * _paginationInfo.columnGap;
        var colWidthCheck = ((_htmlBodyIsVerticalWritingMode ? _$epubHtml[0].scrollHeight : _$epubHtml[0].scrollWidth) - totalGaps) / _paginationInfo.columnCount;
        colWidthCheck = Math.round(colWidthCheck);

        if (colWidthCheck > _paginationInfo.columnWidth)
        {
            console.debug("ADJUST COLUMN");
            console.log(_paginationInfo.columnWidth);
            console.log(colWidthCheck);
            
            _paginationInfo.columnWidth = colWidthCheck;
        }

        _paginationInfo.spreadCount =  Math.ceil(_paginationInfo.columnCount / _paginationInfo.visibleColumnCount);

        if(_paginationInfo.currentSpreadIndex >= _paginationInfo.spreadCount) {
            _paginationInfo.currentSpreadIndex = _paginationInfo.spreadCount - 1;
        }

        if(_deferredPageRequest) {

            //if there is a request for specific page we get here
            openDeferredElement();
        }
        else {

            //we get here on resizing the viewport
            
            onPaginationChanged(self); // => redraw() => showBook(), so the trick below is not needed

            //We do this to force re-rendering of the document in the iframe.
            //There is a bug in WebView control with right to left columns layout - after resizing the window html document
            //is shifted in side the containing div. Hiding and showing the html element puts document in place.
            
            // USE_ADOBE_DRM. This was commented out in readium-shared-js 0.14 but this will cause the same problem on mac 10.8 and 10.9.
            // But if we do it here, we get a blank page on the first load of the book. So we move this code to onviewportresize() handler.
            //_$epubHtml.hide();
            //setTimeout(function() {
              //  _$epubHtml.show();
                //onPaginationChanged(self); // => redraw() => showBook()
            //}, 50);

        }
    }

//    function shiftBookOfScreen() {
//
//        if(_spine.isLeftToRight()) {
//            _$epubHtml.css("left", (_lastViewPortSize.width + 1000) + "px");
//        }
//        else {
//            _$epubHtml.css("right", (_lastViewPortSize.width + 1000) + "px");
//        }
//    }

    function hideBook()
    {
        if (_currentOpacity != -1) return; // already hidden
        
        _currentOpacity = _$epubHtml.css('opacity');
        _$epubHtml.css('opacity', "0");
    }

    function showBook()
    {
        if (_currentOpacity != -1)
        {
            _$epubHtml.css('opacity', _currentOpacity);
        }
        _currentOpacity = -1;
    }

    this.getFirstVisibleElementCfi = function() {

        var contentOffsets = getVisibleContentOffsets();
        return _navigationLogic.getFirstVisibleElementCfi(contentOffsets);
    };

    //USE_ADOBE_DRM
    this.getScreenEndCFI = function() {

        var contentOffsets = getVisibleContentOffsets();
        if(_currentSpineItem)
        {
            var idref = _currentSpineItem.idref;
            var cfi = _navigationLogic.getLastVisibleElementCfi(contentOffsets);
            return new ReadiumSDK.Models.BookmarkData(idref, cfi);
        }
        return new ReadiumSDK.Models.BookmarkData("", "");
    };    

    //USE_ADOBE_DRM
    this.getContentIFrame = function(idref){
        return _$iframe;
    };      

    this.getPaginationInfo = function() {

        var paginationInfo = new ReadiumSDK.Models.CurrentPagesInfo(_spine.items.length, false, _spine.direction);

        if(!_currentSpineItem) {
            return paginationInfo;
        }

        var pageIndexes = getOpenPageIndexes();

        for(var i = 0, count = pageIndexes.length; i < count; i++) {

            paginationInfo.addOpenPage(pageIndexes[i], _paginationInfo.columnCount, _currentSpineItem.idref, _currentSpineItem.index);
        }

        return paginationInfo;

    };

    function getOpenPageIndexes() {

        var indexes = [];

        var currentPage = _paginationInfo.currentSpreadIndex * _paginationInfo.visibleColumnCount;

        for(var i = 0; i < _paginationInfo.visibleColumnCount && (currentPage + i) < _paginationInfo.columnCount; i++) {

            indexes.push(currentPage + i);
        }

        return indexes;

    }

	//USE_ADOBE_DRM
    function _getCSSStyle(a) {
    var epubContentDocument = _$iframe[0].contentDocument;
    var sheets = epubContentDocument.styleSheets, resultObject = {};
    for (var i in sheets) {
        var rules = sheets[i].rules || sheets[i].cssRules;
        for (var r in rules) {
            try {
              if (a.is(rules[r].selectorText)) {
                resultObject = $.extend(resultObject, _cssStyleToJavascriptObject(rules[r].style), _cssStyleToJavascriptObject(a.attr('style')));
                }  
            } catch (err)
            {
                //ignore the error msgs
            }
        }
    }
    return resultObject;
    }

	//USE_ADOBE_DRM
    function _cssStyleToJavascriptObject(css) {
        var s = {};
        if (!css) return s;
        if (css instanceof CSSStyleDeclaration) {
            for (var i in css) {
                if ((css[i]).toLowerCase) {
                    s[(css[i]).toLowerCase()] = (css[css[i]]);
                }
            }
        } else if (typeof css == "string") {
            css = css.split("; ");
            for (var i in css) {
                var l = css[i].split(": ");
                s[l[0].toLowerCase()] = (l[1]);
            }
        }
        else if (typeof css.cssText == "string") {
            var cssString = css.cssText.split(";");
            for (var i in cssString) {
                var l = cssString[i].split(":");

                if(l[0] && l[1]) { //make sure we not try to trip undefined object
                    var key = l[0].trim().toLowerCase();
                    var value = (l[1]).trim()
                    s[key] = value;
                }
            }
        }
        return s;
    }


    //USE_ADOBE_DRM
    function _getScaledSize(sizeToFit,image) {

        var actualImageSize = { width:image[0].naturalWidth, height:image[0].naturalHeight },
            widthDiff = (actualImageSize.width - sizeToFit.width)/actualImageSize.width,
            heightDiff = (actualImageSize.height - sizeToFit.height)/actualImageSize.height;

        var resultSize = {width:0,height:0};

        if(widthDiff > heightDiff) {
            resultSize.width  = sizeToFit.width;
            resultSize.height = actualImageSize.height - actualImageSize.height * widthDiff;
        }else {
            resultSize.height = sizeToFit.height;
            resultSize.width = actualImageSize.width - actualImageSize.width * heightDiff;
        }

        return resultSize;
    };


    //USE_ADOBE_DRM
    function updateImageWrapperSize(viewPortSize) {

        if(!_$epubHtml) {
            return;
        }

        var $elem,$wrapper;


        //_$htmlBody.css("height", viewPortSize.height + "px");
         
        $('img', _$epubHtml).each(function(){
            $elem = $(this);

            var wrapperId = $elem.data('rmImgId');

            if (!_imageOriginalStyleMap[wrapperId]) {
                 return ;//ignore the image has not been resized yet
            }

            $wrapper = $('#'+wrapperId, _$epubHtml);

            //remove all styles from the element
            removeStyles($elem);

            $elem.css(_imageOriginalStyleMap[wrapperId]);
            $elem.css('max-width', '98%');
            $elem.css('max-height', '98%');

            $wrapper.css('width', viewPortSize.width+'px');
            $wrapper.css('height', viewPortSize.height+'px');

            var columnWidth = parseInt(_$epubHtml.css('max-width'),10);
            var columnGap = parseInt(_$epubHtml.css('-webkit-column-gap'),10);

            if (columnGap == undefined || isNaN(columnGap)) {
                columnGap = parseInt(_$epubHtml.css('column-gap'),10);
            }

            var numberOfColumnsInPage = Math.floor(viewPortSize.width/columnWidth);

            var totalGap = columnGap * numberOfColumnsInPage + columnGap;

            var calculatedSize = { width:columnWidth-totalGap, height:parseInt($elem.css('height'),10) };

            removeStyles($elem);

            var sizeToFit = calculatedSize;

            if(calculatedSize.width > viewPortSize.width || calculatedSize.height > viewPortSize.height) {
                calculatedSize = viewPortSize;
            }

            calculatedSize = _getScaledSize(calculatedSize,$elem);
            
            $wrapper = $('#'+wrapperId, _$epubHtml);
            $elem.css('width', calculatedSize.width+'px');
            $elem.css('height', calculatedSize.height+'px');

            $wrapper.css('width', calculatedSize.width+'px');
            $wrapper.css('height', calculatedSize.height+'px');
           
        });
    }


	//USE_ADOBE_DRM
    var removeStyles = function (elem) {

        elem.children().each(function() {
                $(this).removeClass();
                $(this).removeAttr('style');
        });
        elem.removeClass();
        elem.removeAttr('style');
    };

	//USE_ADOBE_DRM
    function isIE () {
      var myNav = navigator.userAgent.toLowerCase();
      return (myNav.indexOf('msie') != -1) ? parseInt(myNav.split('msie')[1]) : false;
    }

	//USE_ADOBE_DRM
    var _imageIdGenerator = 0;
    var _imageOriginalStyleMap = {};

	//USE_ADOBE_DRM
    function wrapImageWithDiv($elem) {
        //rpandian : an id is created and used for wrapping the element with div tag
            var id = "rmImgId_"+ (++_imageIdGenerator);

            _imageOriginalStyleMap[id] = _getCSSStyle($elem);

            //HACK!!!! rpandian wrap the image with a div tag, set dimensions of the div element;
            var wrapString = "<div id='"+id+"' class='rmsdk_ImgWrapper' style='overflow:hidden; width:"+_lastViewPortSize.width+"px ; height:"+_lastViewPortSize.height+"px ;' ></div>";
            $elem.wrap(wrapString);

            //get style text of the image element
            var elemStyle = $.extend({},_imageOriginalStyleMap[id]);

            //IE ignores the height& width and hence the columnization when attribure float is not none
            if(isIE()) { 
                elemStyle['float'] = 'none'
            }

            //copy the css to the wrapper element
            $('#'+id,_$epubHtml).css(elemStyle);

            //remove all styles and classes from the image element
            removeStyles($elem);

            //store the id,calculated in the data attribute of the element
            $elem.data('rmImgId',id);
    }


    //we need this styles for css columnizer not to chop big images
    function resizeImages() {

        if(!_$epubHtml) {
            return;
        }

        var $elem;
        var height;
        var width;

        $('img', _$epubHtml).each(function(){

            $elem = $(this);

			//USE_ADOBE_DRM in IE max-width & max-height css properties are ignored if the immideate container doesnt have width &  height mentioned explicitly
			//we wrap all the image elements with our own div element 
            // wrapImageWithDiv($elem);

            // if we set max-width/max-height to 100% columnizing engine chops images embedded in the text
            // (but not if we set it to 99-98%) go figure.
            $elem.css('max-width', '98%');
            $elem.css('max-height', '98%');

            if(!$elem.css('height')) {
                $elem.css('height', 'auto');
            }

            if(!$elem.css('width')) {
                $elem.css('width', 'auto');
            }

            // updateImageWrapperSize(_lastViewPortSize);//USE_ADOBE_DRM

        });
    }

    this.bookmarkCurrentPage = function() {

        if(!_currentSpineItem) {

            return new ReadiumSDK.Models.BookmarkData("", "");
        }

        return new ReadiumSDK.Models.BookmarkData(_currentSpineItem.idref, self.getFirstVisibleElementCfi());
    };

    function getVisibleContentOffsets() {
        //TODO: _htmlBodyIsVerticalWritingMode ? (_lastViewPortSize.height * _paginationInfo.currentSpreadIndex)
        // NOT used with options.rectangleBased anyway (see CfiNavigationLogic constructor call, here in this reflow engine class)
        var columnsLeftOfViewport = Math.round(_paginationInfo.pageOffset / (_paginationInfo.columnWidth + _paginationInfo.columnGap));

        var topOffset =  columnsLeftOfViewport * _$contentFrame.height();
        var bottomOffset = topOffset + _paginationInfo.visibleColumnCount * _$contentFrame.height();

        return {top: topOffset, bottom: bottomOffset};
    }

    this.getLoadedSpineItems = function() {
        return [_currentSpineItem];
    };

    this.getElementByCfi = function(spineItem, cfi, classBlacklist, elementBlacklist, idBlacklist) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElementByCfi(cfi, classBlacklist, elementBlacklist, idBlacklist);
    };

    this.getElementById = function(spineItem, id) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElementById(id);
    };

    this.getElement = function(spineItem, selector) {

        if(spineItem != _currentSpineItem) {
            console.error("spine item is not loaded");
            return undefined;
        }

        return _navigationLogic.getElement(selector);
    };

    this.getFirstVisibleMediaOverlayElement = function() {

        var visibleContentOffsets = getVisibleContentOffsets();
        return _navigationLogic.getFirstVisibleMediaOverlayElement(visibleContentOffsets);
    };
    
    // /**
    //  * @deprecated
    //  */
    // this.getVisibleMediaOverlayElements = function() {
    // 
    //     var visibleContentOffsets = getVisibleContentOffsets();
    //     return _navigationLogic.getVisibleMediaOverlayElements(visibleContentOffsets);
    // };

    this.insureElementVisibility = function(spineItemId, element, initiator) {

        var $element = $(element);
        if(_navigationLogic.isElementVisible($element, getVisibleContentOffsets()))
        {
            return;
        }

        var page = _navigationLogic.getPageForElement($element);

        if(page == -1)
        {
            return;
        }

        var openPageRequest = new ReadiumSDK.Models.PageOpenRequest(_currentSpineItem, initiator);
        openPageRequest.setPageIndex(page);

        var id = element.id;
        if (!id)
        {
            id = element.getAttribute("id");
        }
        
        if (id)
        {
            openPageRequest.setElementId(id);
        }

        self.openPage(openPageRequest);
    }

};
